﻿using UnityEngine;
using System;
using System.Reflection;
using System.Diagnostics;
using System.Collections.Generic;
using System.Runtime.Serialization.Formatters.Binary;

namespace Ld31
{
    public class XInvokeProfile
    {
        private delegate void XInvokeDelegate();

        //public float Delay;
        public float Interval;
        public float Last;
        public string Name;
        public object Target;
        private int _counter;
        private float _timestamp;
        private XInvokeDelegate method;

        public int Counter
        {
            get
            {
                return _counter;
            }
        }

        /// <summary>
        /// Defines the invocation of a methodName on a target object
        /// </summary>
        public XInvokeProfile(object target, float delay, float interval, float timestamp, string methodName)
        {
            /*            if (interval < 0.015f)
            {
                throw new ArgumentException("Interval must be > 0.015");
            }*/

            if (target == null)
            {
                throw new NullReferenceException("Invoke target cannot be null");
            }

            if (methodName.Length == 0)
            {
                throw new ArgumentException("Invoke methodName cannot be empty");
            }

            MethodInfo mInfo = target.GetType().GetMethod(methodName, BindingFlags.Instance | BindingFlags.Public | BindingFlags.NonPublic);

            if (mInfo == null)
            {
                throw new ArgumentException("Invoke method " + methodName + " was not found in " + target.GetType());
            }

            method = (XInvokeDelegate)Delegate.CreateDelegate(
                typeof(XInvokeDelegate),
                target,
                mInfo
            );

            //Delay = delay;
            Interval = interval;
            Name = methodName;
            Last = timestamp - interval + delay;
            Target = target;
            _timestamp = timestamp;
            _counter = 0;
            #if DEBUGTHIS
            Debug.Log("XInvokeProfile Created: " + t.GetType() + " -> " + Name + ", " + Delay + ", " + Interval);
            #endif
        }

        public float Timestamp
        {
            get
            {
                return _timestamp;
            }
        }

        public void Execute()
        {
            method();
            _counter++;
        }
    }

    public class Clock : MonoBehaviour
    {

        /* Maintains a custom game Clock and controls game speed */

        public float MinSpeed = 1f;
        public float MaxSpeed = 16f;

        [SerializeField] private float _gameSpeed = 1f;
        [SerializeField] private float _gameSpeedInverse = 1f;
        [SerializeField] private float _gameTime = 0;

        private const float INTERVAL_CUTOFF = 0.01f;

        private List<XInvokeProfile> XInvokeRepeatingQueue = new List<XInvokeProfile>();

        void Start()
        {
            //InvokeRepeating("NextStep", 0f, _gameSpeed);
        }

        void Update()
        {
            _gameTime += Time.deltaTime * _gameSpeed;
            XInvokeTick();
        }

        /// <summary>
        /// Game time step count since launch.
        /// </summary>
        public double Step
        {
            get
            {
                return Mathf.Floor(_gameTime);
            }
        }

        /// <summary>
        /// Current speed as multiple of normal speed.
        /// </summary>
        public float Speed
        {
            get
            {
                return _gameSpeed;
            }
            set
            {
                _gameSpeed = Mathf.Clamp(value, MinSpeed, MaxSpeed);
                _gameSpeedInverse = 1 / _gameSpeed;
                //CancelInvoke("NextStep");
                //InvokeRepeating("NextStep", 0f, _gameSpeedInverse);
            }
        }

        /// <summary>
        /// Current game time.
        /// </summary>
        public float Now
        {
            get
            {
                return _gameTime;
            }
        }

        /// <summary>
        /// Increases game time speed.
        /// </summary>
        public void Faster()
        {
            Speed = _gameSpeed * 2f;
            UnityEngine.Debug.Log("Game Speed Increased, speed is now " + Speed + "x");
        }

        /// <summary>
        /// Reduces game time speed.
        /// </summary>
        public void Slower()
        {
            Speed = _gameSpeed * 0.5f;
            UnityEngine.Debug.Log("Game Speed Reduced, speed is now " + Speed + "x");
        }

        /// <summary>
        /// Current speed as fraction of normal speed.
        /// </summary>
        public float Fraction
        {
            get
            {
                return _gameSpeedInverse;
            }
        }

        private void XInvokeTick()
        {
            if (XInvokeRepeatingQueue.Count == 0)
                return;

            // remove profiles that have lost their target for some reason (destroyed).
            XInvokeRepeatingQueue.RemoveAll(k => k.Target == null);

            for (int index = 0; index < XInvokeRepeatingQueue.Count; index++)
            {
                XInvokeProfile each = XInvokeRepeatingQueue[index];

                if (each.Last + each.Interval < Now)
                {
                    #if DEBUGTHIS
                    Debug.Log(GetType() + "-" + each.Name + ": " + each.Last + " + " + each.Interval + " > " + Now);
                    #endif
                    if (each.Target != null)
                    {
                        each.Last = Now;
                        Stopwatch timer = new Stopwatch();
                        timer.Start();
                        each.Execute();
                        timer.Stop();
                        // UnityEngine.Debug.Log(each.Target + "-" + each.Name + ": " + timer.Elapsed + "s");
                    }

                    // remove if non-repeating
                    if (each.Interval < INTERVAL_CUTOFF)
                        XInvokeRepeatingQueue.RemoveAt(index);
                }
            }
        }

        /// <summary>
        /// Clears all queued invocations referencing a target object.
        /// </summary>
        public void XCancelInvoke(object target)
        {
            XInvokeRepeatingQueue.RemoveAll(item => item.Target.Equals(target));
            #if DEBUGTHIS
            Debug.Log("CancelInvoke: " + target.GetType() + " -> *");
            #endif
        }

        /// <summary>
        /// Removes a methodName for a given target object from the invoke-queue.
        /// </summary>
        public void XCancelInvoke(object target, string methodName)
        {
            var index = XInvokeRepeatingQueue.FindIndex(item => item.Name.Equals(methodName) && item.Target.Equals(target));
            if (index >= 0)
            {
                #if DEBUGTHIS
                Debug.Log("CancelInvoke: " + target.GetType() + " -> " + methodName);
                #endif
                XInvokeRepeatingQueue.RemoveAt(index);
            }
        }

        /// <summary>
        /// Checks if a methodName on target is queued to be invoked.
        /// </summary>
        public bool XIsInvoking(object target, string methodName)
        {
            if (XInvokeRepeatingQueue.Count == 0)
                return false;

            foreach (XInvokeProfile each in XInvokeRepeatingQueue)
            {
                if (each.Name.Equals(methodName) && each.Target.Equals(target))
                    return true;
            }
            return false;
        }

        /// <summary>
        /// Calls the given methodName on the target object repeatedly. Each target/method pair can 
        /// only exist once in the repeater queue. Repeated calls for the same target/method are 
        /// ignored. To change an invoke use XCancelInvoke to delete it, then recreate it witch 
        /// changed parameters. Interval and delay are subject to Clock.Speed.
        /// </summary>
        public void XInvoke(object target, string methodName, float delay)
        {
            if (XIsInvoking(target, methodName))
            {
                #if DEBUGTHIS
                Debug.Log("Method " + methodName + " already invoking.");
                #endif
                return;
            }

            XInvokeRepeatingQueue.Add(
                new XInvokeProfile(
                    target,
                    delay,
                    0,
                    Now,
                    methodName
                )
            );
            UnityEngine.Debug.Log("-------- Invoke: " + target.GetType().Name + " -> " + methodName);
        }

        /// <summary>
        /// Calls the given methodName on the target object repeatedly. Each target/method pair can 
        /// only exist once in the repeater queue. Repeated calls for the same target/method are 
        /// ignored. To change an invoke use XCancelInvoke to delete it, then recreate it witch 
        /// changed parameters. Interval and delay are subject to Clock.Speed.
        /// </summary>
        public void XInvokeRepeating(object target, string methodName, float delay, float interval)
        {
            if (interval < INTERVAL_CUTOFF)
                throw new ArgumentException("interval must be > " + INTERVAL_CUTOFF);

            if (XIsInvoking(target, methodName))
            {
                #if DEBUGTHIS
                Debug.Log("Method " + methodName + " already invoking.");
                #endif
                return;
            }

            XInvokeRepeatingQueue.Add(
                new XInvokeProfile(
                    target,
                    delay,
                    interval,
                    Now,
                    methodName
                )
            );
            UnityEngine.Debug.Log("-------- Invoke Repeating: " + target.GetType().Name + " -> " + methodName);
        }
    }
}