﻿using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System;

namespace Ld31
{
    /// <summary>
    /// Maintains a list of selected entities.
    /// </summary>
    public class EntitySelection : Root, IEnumerable
    {
        static public event EventHandler OnSet;
        static public event EventHandler OnClear;
        static public event EventHandler OnAdd;
        static public event EventHandler OnRemove;

        private List<GameObject> _selection = new List<GameObject>();

        private void Set(GameObject go)
        {
            _selection.Clear();
            _selection.Add(go);
            if (OnSet != null)
                OnSet(this, EventArgs.Empty);
            Debug.Log(Count());

        }

        private void Add(GameObject go)
        {
            if (!Contains(go))
            {
                _selection.Add(go);
                if (OnAdd != null)
                    OnAdd(this, EventArgs.Empty);
            }
        }

        private void Remove(GameObject go)
        {
            _selection.Remove(go);
            if (OnRemove != null)
                OnRemove(this, EventArgs.Empty);
        }

        public bool Contains(GameObject go)
        {
            if (go != null && _selection.Count > 0)
                return _selection.Contains(go);
            else
                return false;
        }

        public int Count()
        {
            return _selection.Count;
        }

        public void Clear()
        {
            _selection.Clear();
            if (OnClear != null)
                OnClear(this, EventArgs.Empty);

        }

        public IEnumerator GetEnumerator()
        {
            return _selection.GetEnumerator();
        }

        public void Try(Camera camera, Vector3 pointA, Vector3 pointB)
        {
            var worldPointA = camera.ScreenToWorldPoint(pointA);
            var worldPointB = camera.ScreenToWorldPoint(pointB);
            var startX = (int)Mathf.Min(worldPointA.x, worldPointB.x);
            var startY = (int)Mathf.Min(worldPointA.y, worldPointB.y);
            var endX = (int)Mathf.Max(worldPointA.x, worldPointB.x);
            var endY = (int)Mathf.Max(worldPointA.y, worldPointB.y);

            if (!Game.Controls.GetShift())
                Clear();

            for (int y = startY; y <= endY; y++)
                for (int x = startX; x <= endX; x++)
                    if (Game.Map._Data[x, y].Structure != null)
                        Add(Game.Map._Data[x, y].Structure.gameObject);
        }

        /// <summary>
        /// Probes the point in camera-space for selectable entities (Locations, Agents) and adds them to the 
        /// selection. Click into null space clears the selection. Shift-Click allows multiple selection.
        /// </summary>
        public void Try(Camera camera, Vector3 screenPoint)
        {
            // define a ray
            Ray ray = camera.ScreenPointToRay(screenPoint);
            Debug.DrawRay(ray.origin, ray.direction * 108, Color.white, 5f, false);

            // collide the ray with entities we like to be selectable
            var layers = (
                1 << LayerMask.NameToLayer("Structures")
            );
            RaycastHit2D hitInfo = Physics2D.GetRayIntersection(ray, 100f, layers);

            if (Game.Controls.GetShift())
            {
                // if something was hit, add it to the selection
                if (hitInfo.collider != null)
                {
                    if (Contains(hitInfo.collider.gameObject))
                        Remove(hitInfo.collider.gameObject);
                    else
                        Add(hitInfo.collider.gameObject);
                }

            }
            else
            {
                if (hitInfo.collider != null)
                    Set(hitInfo.collider.gameObject);
                else
                    Clear();
            }
        }
    }
}